/*
 * MSU CUDA Course Examples and Exercises.
 *
 * Copyright (c) 2011 Dmitry Mikushin
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising 
 * from the use of this software.
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it freely,
 * without any restrictons.
 */

#include <assert.h>
#include <cuda.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

// Maximum allowed results difference.
#define EPS 1e-04

// Draw or don't draw field in case if
// difference > EPS.
#define DRAWDIFFS 0

// Grid block size (see comment in main).
#define BLOCK_LENGTH 32
#define BLOCK_HEIGHT 16

#define IN(i,j) in[i + (j) * nx]
#define OUT(i,j) out[i + (j) * nx]

// Reference for function performing simple
// results difference visualization.
extern "C" void draw_diffs(
	int nx, int ny, float* diffs, float eps);

// GPU device kernel.
__global__ void pattern2d_gpu(
	int bx, int nx, int ex, int by, int ny, int ey,
	float* in, float* out)
{
	// Compute absolute (i,j) indexes for
	// the current GPU thread using grid mapping params.
	int i = blockIdx.x * BLOCK_LENGTH + threadIdx.x + bx;
	int j = blockIdx.y * BLOCK_HEIGHT + threadIdx.y + by;
	
	// Compute one data point - a piece of
	// work for the current GPU thread.
	OUT(i,j) = 0.1f + sqrtf(fabs(IN(i,j) + IN(i-1,j) + IN(i+1,j) -
		2.0f * IN(i,j-1) + 3.0f * IN(i,j+1)));
}

// CPU control implementation.
void pattern2d_cpu(
	int bx, int nx, int ex, int by, int ny, int ey,
	float* in, float* out)
{
	for (int j = by; j < ny - ey; j++)
		for (int i = bx; i < nx - ex; i++)
			OUT(i,j) = sqrtf(fabs(IN(i,j) + IN(i-1,j) + IN(i+1,j) -
				2.0f * IN(i,j-1) + 3.0f * IN(i,j+1)));
}

int usage(char* name)
{
	printf("Sample 2D field processing\n");
	printf("Usage: %s <nx> <ny>\n", name);
	return 0;
}

int main(int argc, char* argv[])
{
	if (argc != 3) return usage(argv[0]);
	
	int nx = atoi(argv[1]);
	int ny = atoi(argv[2]);
	long np = nx * ny;
	
	if ((nx <= 0) || (ny <= 0)) return usage(argv[0]);

	size_t size = sizeof(float) * np;

	// Generate CPU data input array according
	// to the specified X and Y dimensions.
	float* in_cpu = (float*)malloc(size);
	float invdrandmax = 1.0 / RAND_MAX;
	for (long i = 0; i < np; i++)
		in_cpu[i] = rand() * invdrandmax;
	
	// Create GPU data array and copy previously
	// generated data to it.
	float* in_gpu;
	cudaError_t status = cudaMalloc((void**)&in_gpu, size);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot allocate device input data array, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	status = cudaMemcpy(in_gpu, in_cpu, size, cudaMemcpyHostToDevice);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot copy input data to device, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	
	// Create CPU data output array and get
	// control result using CPU function.
	float* control = (float*)malloc(size);
	memset(control, 0, size);
	pattern2d_cpu(
		1, nx, 1, 1, ny, 1, in_cpu, control);
	
	// Configure GPU computational grid:
	// nx = nblocks_x * block_length
	// ny = nblocks_y * block_height
	//
	// NOTE: we have degree of freedom in
	// selecting how real problem grid maps onto
	// computational grid. Usually these params
	// are tuned to get optimal performance.
	//
	// NOTE: chose of grid/block config is
	// also limited by device properties:
	// - Maximum number of threads per block (512)
	// - Maximum sizes of each dimension of a block (512 x 512 x 64)
	// - Maximum sizes of each dimension of a grid (65535 x 65535 x 1)
	int nblocks_x = (nx - 2) / BLOCK_LENGTH;
	int nblocks_y = (ny - 2) / BLOCK_HEIGHT;
	
	// Perform the same processing on GPU,
	// returning result to GPU array.
	float* out_gpu;
	status = cudaMalloc((void**)&out_gpu, size);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot allocate device output data array, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	status = cudaMemset(out_gpu, 0, size);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot zero device output data aray, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	pattern2d_gpu<<<
		dim3(nblocks_x, nblocks_y, 1),
		dim3(BLOCK_LENGTH, BLOCK_HEIGHT, 1)>>>(
			1, nx, 1, 1, ny, 1, in_gpu, out_gpu);
	status = cudaGetLastError();
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot launch pattern2d on device, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	status = cudaThreadSynchronize();
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot synchronize thread, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
		
	// Copy GPU result from GPU memory to CPU buffer.
	float* out_cpu = (float*)malloc(size);
	status = cudaMemcpy(out_cpu, out_gpu, size, cudaMemcpyDeviceToHost);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot copy output data from device, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}

	// Don't bother with processing the remainder
	// on GPU. Do it on CPU instead.
	pattern2d_cpu(
		1, nx, 1,
		ny - (ny - 2) % BLOCK_HEIGHT - 2, ny, 1,
		in_cpu, out_cpu);
	pattern2d_cpu(
		nx - (nx - 2) % BLOCK_LENGTH - 2, nx, 1,
		1, ny, 1,
		in_cpu, out_cpu);
	
	// Compare results and find the maximum abs difference.
	int maxi = 0, maxj = 0;
	float maxdiff = fabs(out_cpu[0] - control[0]);
	float* diffs = (float*)malloc(size);
	memset(diffs, 0, size);
	for (int j = 0; j < ny; j++)
		for (int i = 0; i < nx; i++)
		{
			float diff = fabs(
				out_cpu[i + j * nx] -
				control[i + j * nx]);
			if (diff > maxdiff)
			{
				maxdiff = diff;
				maxi = i; maxj = j;
			}
			diffs[i + j * nx] = diff;
		}

	printf("Done! abs max diff = %f @ (%d,%d)\n",
		maxdiff, maxi, maxj);

	// If there is results difference,
	// vizualize it.
	if (DRAWDIFFS && (maxdiff > EPS))
	{
		draw_diffs(nx, ny, diffs, EPS);
		printf("Wrote diffs to pattern2d.png\n");
	}
	
	// Release data arrays.
	free(in_cpu);
	status = cudaFree(in_gpu);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free device input data array, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	free(control);
	status = cudaFree(out_gpu);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free device output data array, status = %s\n",
			cudaGetErrorString(status));
		return 1;
	}
	free(out_cpu);
	free(diffs);

	return 0;
}

