/*
 * MSU CUDA Course Examples and Exercises.
 *
 * Copyright (c) 2011 Dmitry Mikushin
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising 
 * from the use of this software.
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it freely,
 * without any restrictons.
 */

#include "pattern2d.h"

#include <cuda_runtime.h>
#include <malloc.h>
#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main(int argc, char* argv[])
{
	// Get problem size from the command line.
	if (argc != 4)
	{
		printf("Usage: %s <nprocesses> <n> <npasses>\n", argv[0]);
		return 0;
	}

	int nprocesses = atoi(argv[1]);
	int n = atoi(argv[2]);
	int npasses = atoi(argv[3]);
	size_t size = n * n * sizeof(float);

	if ((nprocesses <= 0) || (n <= 0) || (npasses <= 0)) return 0;

	int ndevices = 0;
	cudaError_t cuda_status = cudaGetDeviceCount(&ndevices);
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot get the cuda device count by process 0, status = %s\n",
			cudaGetErrorString(cuda_status));
		return 1;
	}

	// Create device input & output buffers for all devices.
	float** din1 = (float**)malloc(sizeof(float*) * nprocesses);
	float** din2 = (float**)malloc(sizeof(float*) * nprocesses);
	float** dout = (float**)malloc(sizeof(float*) * nprocesses);
	for (int iprocess = 0; iprocess < nprocesses; iprocess++)
	{
		cuda_status = cudaSetDevice(iprocess);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot set CUDA device %d, status= %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
		cudaMalloc((void**)(&din1[iprocess]), size);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot allocate input device buffer for device %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
		cudaMalloc((void**)(&din2[iprocess]), size);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot allocate input device buffer for device %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
		cudaMalloc((void**)(&dout[iprocess]), size);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot allocate output device buffer by process %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
	}

	// Initialize MPI. From this point the specified
	// number of processes will be executed in parallel.
	int mpi_status = MPI_Init(&argc, &argv);
	int mpi_error_msg_length;
	char mpi_error_msg[MPI_MAX_ERROR_STRING];
	if (mpi_status != MPI_SUCCESS)
	{
		MPI_Error_string(mpi_status, mpi_error_msg, &mpi_error_msg_length);
		fprintf(stderr, "Cannot initialize MPI, status = %s\n",
			mpi_error_msg);
		return 1;
	}
	
	// Get the rank (index) of the current MPI process
	// in the global communicator.
	int iprocess;
	mpi_status = MPI_Comm_rank(MPI_COMM_WORLD, &iprocess);
	if (mpi_status != MPI_SUCCESS)
	{
		MPI_Error_string(mpi_status, mpi_error_msg, &mpi_error_msg_length);
		fprintf(stderr, "Cannot retrieve the rank of current MPI process, status = %s\n",
			mpi_error_msg);
		return 1;
	}

	nprocesses = atoi(argv[1]);
	n = atoi(argv[2]);
	npasses = atoi(argv[3]);
	size = n * n * sizeof(float);
	
	if ((n <= 0) || (npasses <= 0)) return 0;

        // Assign unique device to each MPI process.
	cuda_status = cudaSetDevice(iprocess);
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot set CUDA device %d, status= %s\n",
			iprocess, cudaGetErrorString(cuda_status));
		return 1;
	}
	
	float* hin = (float*)malloc(size);
	float* hout = (float*)malloc(size);
	
	// Generate random input data.
	double dinvrmax = 1.0 / RAND_MAX;
	for (int i = 0; i < n * n; i++)
	{
		for (int j = 0; j < iprocess + 1; j++)
			hin[i] += rand() * dinvrmax;
		hin[i] /= iprocess + 1;
	}
	
	// Copy input data generated on host to device buffer.
	cuda_status = cudaMemcpy(din1[iprocess], hin, size, cudaMemcpyHostToDevice); 
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot copy input data from host to device by process %d, status = %s\n",
			iprocess, cudaGetErrorString(cuda_status));
		return 1;
	}
	
	// Perform the specified number of processing passes.
	for (int ipass = 0; ipass < npasses; ipass++)
	{
		// Fill output device buffer will zeros.
		cuda_status = cudaMemset(dout[iprocess], 0, size);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot fill output device buffer with zeros by process %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}

		// Process data on GPU.
		pattern2d_gpu(1, n, 1, 1, n, 1, din1[iprocess], dout[iprocess]);

		// Wait for GPU kernels to finish processing.
		cuda_status = cudaDeviceSynchronize();
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot synchronize GPU kernel by process %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
	
		// Copy output data back from device to host.
		cuda_status = cudaMemcpy(hout, dout[iprocess], size, cudaMemcpyDeviceToHost); 
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot copy output data from device to host by process %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
	
		// Output average value of the resulting field.
		float avg = 0.0;
		for (int i = 0; i < n * n; i++)
			avg += hout[i];
		avg /= n * n;
		
		int iprev = iprocess - 1; iprev += (iprev < 0) ? nprocesses : 0;

		// Recieve input device buffer from input device buffer of the previous
		// process, using peer-to-peer memcpy between GPUs.
		printf("Process %d result avg = %f, receiving new input to %p from %p of process %d\n",
			iprocess, avg, din2[iprocess], din1[iprev], iprev);
		
		cuda_status = cudaMemcpyPeer(din2[iprocess], iprocess, din1[iprev], iprev, size);
		if (cuda_status != cudaSuccess)
		{
			fprintf(stderr, "Cannot copy GPU data in peer-to-peer mode by process %d, status = %s\n",
				iprocess, cudaGetErrorString(cuda_status));
			return 1;
		}
		cuda_status = cudaDeviceSynchronize();
		MPI_Barrier(MPI_COMM_WORLD);

                cuda_status = cudaMemcpy(hout, din2[iprocess], size, cudaMemcpyDeviceToHost);
                if (cuda_status != cudaSuccess)
                {
                        fprintf(stderr, "Cannot copy output data from device to host by process %d, status = %s\n",
                                iprocess, cudaGetErrorString(cuda_status));
                        return 1;
                }

                // Output average value of the resulting field.
                avg = 0.0;
                for (int i = 0; i < n * n; i++)
                        avg += hout[i];
                avg /= n * n;

		printf("Process %d received buffer avg = %f to %p\n", iprocess, avg, din2[iprocess]);

		MPI_Barrier(MPI_COMM_WORLD);

		// Swap buffers.
		//float* swap1 = din1[iprocess]; din1[iprocess] = din2[iprocess]; din2[iprocess] = swap1;
		//float* swap2 = din1[iprev]; din1[iprev] = din2[iprev]; din2[iprev] = swap2;
	}
	
	cuda_status = cudaFree(din1[iprocess]);
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free input device buffer by process %d, status = %d\n",
			iprocess, cudaGetErrorString(cuda_status));
		return 1;
	}
	cuda_status = cudaFree(din2[iprocess]);
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free input device buffer by process %d, status = %d\n",
			iprocess, cudaGetErrorString(cuda_status));
		return 1;
	}
	
	cuda_status = cudaFree(dout[iprocess]);
	if (cuda_status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free output device buffer by process d, status = %d\n",
			iprocess, cudaGetErrorString(cuda_status));
		return 1;
	}

	free(hin);
	free(hout);

	mpi_status = MPI_Finalize();
	if (mpi_status != MPI_SUCCESS)
	{
		MPI_Error_string(mpi_status, mpi_error_msg, &mpi_error_msg_length);
		fprintf(stderr, "Cannot finalize MPI, status = %s\n",
			mpi_error_msg);
		return 1;
	}
	
	return 0;
}

