/*
 * MSU CUDA Course Examples and Exercises.
 *
 * Copyright (c) 2011 Dmitry Mikushin
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising 
 * from the use of this software.
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it freely,
 * without any restrictons.
 */

#include "pattern2d.h"

#include <cuda_runtime.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SZPAGE 4096

int main(int argc, char* argv[])
{
	if (argc != 2)
	{
		printf("Usage: %s <n>\n", argv[0]);
		return 0;
	}

	int count = 0;
	cudaError_t status = cudaGetDeviceCount(&count);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot get CUDA devices count, status = %d\n", status);
		return 1;
	}
	if (!count)
	{
		fprintf(stderr, "No CUDA-enabled devices found\n");
		return 1;
	}

	status = cudaSetDeviceFlags(cudaDeviceMapHost);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot set device flags, status = %d\n", status);
		return 1;
	}

	int n = atoi(argv[1]);
	size_t size = n * n * sizeof(float);

	// Assume we have external unaligned allocation, that needs to be
	// passed to device. Certainly, it's better to alloc directly with
	// cudaMallocHost, then there is no need of all this aligning crap.
	float* din = (float*)malloc(size);
	size_t din_shift = (size_t)din % SZPAGE;
	float* adin = (float*)((char*)din - din_shift);
	
	// Compute aligned size: account overheads to page left and right borders.
	size_t inasize = size;
	if (size % SZPAGE)
		inasize -= size % 4096 + 4096;
	if (((size_t)din % 4096) && ((size + (size_t)din % 4096) % 4096))
		inasize += 4096;

	status = cudaHostRegister(adin, inasize, cudaHostRegisterMapped);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot register input device buffer, status = %d\n", status);
		return 1;
	}

	float* mdin = din;
	status = cudaHostGetDevicePointer((void**)&mdin, adin, 0);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot map input buffer to device memory, status = %d\n", status);
		return 1;
	}
	mdin = (float*)((char*)mdin + din_shift);
	
	float* dout = (float*)malloc(size);
	size_t dout_shift = (size_t)dout % 4096;
	float* adout = (float*)((char*)dout - dout_shift);

	// Compute aligned size: account overheads to page left and right borders.
	size_t outasize = size;
	if (size % SZPAGE)
		outasize -= size % 4096 + 4096;
	if (((size_t)dout % 4096) && ((size + (size_t)dout % 4096) % 4096))
		outasize += 4096;

	status = cudaHostRegister(adout, outasize, cudaHostRegisterMapped);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot register output device buffer, status = %d\n", status);
		return 1;
	}

	float* mdout = dout;
	status = cudaHostGetDevicePointer((void**)&mdout, adout, 0);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot map output buffer to device memory, status = %d\n", status);
		return 1;
	}
	mdout = (float*)((char*)mdout + dout_shift);

	float* hin = (float*)malloc(size);
	float* hout = (float*)malloc(size);

	// Fill output will zeros (actually, to zero the borders).
	memset(dout, 0, size);
	memset(hout, 0, size);
	
	// Generate random input data.
	double dinvrmax = 1.0 / RAND_MAX;
	for (int i = 0; i < n * n; i++)
	{
		din[i] = rand() * dinvrmax;
		hin[i] = din[i];
	}
	
	// Process data on CPU.
	pattern2d_cpu(1, n, 1, 1, n, 1, hin, hout);

	// Process data on GPU.
	pattern2d_gpu(1, n, 1, 1, n, 1, mdin, mdout);

	// Wait for GPU kernels to finish processing.
	status = cudaThreadSynchronize();
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot synchronize GPU kernels, status = %d\n", status);
		return 1;
	}
	
	// Compare results.
	float maxabsdiff = 0.0;
	for (int i = 0; i < n * n; i++)
	{
		float absdiff = fabs(dout[i] - hout[i]);
		maxabsdiff = fmax(maxabsdiff, absdiff);
	}
	printf("Max abs diff = %f\n", maxabsdiff);
	
	status = cudaHostUnregister(adin);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot unregister input device buffer, status = %d\n", status);
		return 1;
	}
	free(din);
	
	status = cudaHostUnregister(adout);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot unregister output device buffer, status = %d\n", status);
		return 1;
	}
	free(dout);

	free(hin);
	free(hout);
	
	return 0;
}

