/*
 * MSU CUDA Course Examples and Exercises.
 *
 * Copyright (c) 2011 Dmitry Mikushin
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising 
 * from the use of this software.
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it freely,
 * without any restrictons.
 */

#include "pattern2d.h"

#include <cuda_runtime.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main(int argc, char* argv[])
{
	if (argc != 2)
	{
		printf("Usage: %s <n>\n", argv[0]);
		return 0;
	}

	int count = 0;
	cudaError_t status = cudaGetDeviceCount(&count);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot get CUDA devices count, status = d\n", status);
		return 1;
	}
	if (!count)
	{
		fprintf(stderr, "No CUDA-enabled devices found\n");
		return 1;
	}
	
	int n = atoi(argv[1]);
	size_t size = n * n * sizeof(float);

	float* din;
	status = cudaMallocHost((void**)&din, size);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot allocate input device buffer, status = %d\n", status);
		return 1;
	}
	
	float* dout;
	status = cudaMallocHost((void**)&dout, size);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot allocate output device buffer, status = %d\n", status);
		return 1;
	}

	float* hin = (float*)malloc(size);
	float* hout = (float*)malloc(size);

	// Fill output will zeros (actually, to zero the borders).
	memset(dout, 0, size);
	memset(hout, 0, size);
	
	// Generate random input data.
	double dinvrmax = 1.0 / RAND_MAX;
	for (int i = 0; i < n * n; i++)
	{
		din[i] = rand() * dinvrmax;
		hin[i] = din[i];
	}
	
	// Process data on CPU.
	pattern2d_cpu(1, n, 1, 1, n, 1, hin, hout);

	// Process data on GPU.
	pattern2d_gpu(1, n, 1, 1, n, 1, din, dout);

	// Wait for GPU kernels to finish processing.
	status = cudaThreadSynchronize();
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot synchronize GPU kernels, status = %d\n", status);
		return 1;
	}
	
	// Compare results.
	float maxabsdiff = 0.0;
	for (int i = 0; i < n * n; i++)
	{
		float absdiff = fabs(dout[i] - hout[i]);
		maxabsdiff = fmax(maxabsdiff, absdiff);
	}
	printf("Max abs diff = %f\n", maxabsdiff);
	
	status = cudaFreeHost(din);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free input device buffer, status = %d\n", status);
		return 1;
	}
	
	status = cudaFreeHost(dout);
	if (status != cudaSuccess)
	{
		fprintf(stderr, "Cannot free output device buffer, status = %d\n", status);
		return 1;
	}

	free(hin);
	free(hout);
	
	return 0;
}

